"""
Exceptions that are raised at various parts of this library.
"""

from simple_openid_connect.data import AuthenticationErrorResponse


class OpenidProtocolError(Exception):
    """
    A generic error that is raised when the OpenID protocol was irrecoverably violated
    """

    def __init__(self, msg: str, *data) -> None:
        super().__init__(msg, *data)


class AuthenticationFailedError(Exception):
    """
    A previous authentication attempt has failed
    """

    def __init__(self, error: AuthenticationErrorResponse):
        super().__init__(error)



from simple_openid_connect.client_authentication import ClientSecretBasicAuth
from simple_openid_connect.flows import authorization_code_flow
from simple_openid_connect.flows.authorization_code_flow import (
    AuthenticationSuccessResponse,
)


def test_authorization_request(dummy_ua, dummy_openid_provider, dummy_app_server):
    # act
    url = authorization_code_flow.start_authentication(
        dummy_openid_provider.endpoints["authorization"],
        "openid",
        dummy_openid_provider.test_client_id,
        dummy_app_server.callback_url,
    )
    response = dummy_ua.naviagte_to(url)

    # assert
    response_msg = authorization_code_flow.AuthenticationSuccessResponse.parse_url(
        response.url
    )
    assert response_msg.code is not None
    assert response_msg.code != ""


def test_handle_authentication_result(
    dummy_ua, dummy_openid_provider, dummy_app_server
):
    # arrange
    url = authorization_code_flow.start_authentication(
        dummy_openid_provider.endpoints["authorization"],
        "openid",
        dummy_openid_provider.test_client_id,
        dummy_app_server.callback_url,
    )
    response = dummy_ua.naviagte_to(url)

    # act
    response = authorization_code_flow.handle_authentication_result(
        response.url,
        dummy_openid_provider.endpoints["token"],
        ClientSecretBasicAuth(
            dummy_openid_provider.test_client_id,
            dummy_openid_provider.test_client_secret,
        ),
    )

    # assert
    assert response.access_token
    assert response.id_token


def test_token_exchange(dummy_ua, dummy_openid_provider, dummy_app_server):
    # act
    response = authorization_code_flow.exchange_code_for_tokens(
        dummy_openid_provider.endpoints["token"],
        AuthenticationSuccessResponse(code=dummy_openid_provider.cheat_code),
        dummy_app_server.callback_url,
        ClientSecretBasicAuth(
            dummy_openid_provider.test_client_id,
            dummy_openid_provider.test_client_secret,
        ),
    )

    # assert
    assert response.access_token
    assert response.id_token



import logging
import random
import string
from typing import Dict, List, Tuple

import pytest
import requests
import responses
from furl import furl
from requests import PreparedRequest

from simple_openid_connect.data import ProviderMetadata
from simple_openid_connect.flows.authorization_code_flow import (
    AuthenticationRequest,
    AuthenticationSuccessResponse,
    TokenErrorResponse,
    TokenRequest,
    TokenSuccessResponse,
)

logger = logging.getLogger(__name__)


def rand_str() -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=16))


class DummyOpenidProvider:
    iss = "https://provider.example.com/openid-connect"
    endpoints = {
        "authorization": f"{iss}/client_auth",
        "token": f"{iss}/token",
    }

    test_client_id = "test-client"
    test_client_secret = "foobar123"
    test_client_redirect_uri = "https://app.example.com/auth/callback"

    cheat_code = "master access code which can always be exchanged for tokens"

    _valid_auth_codes: List[str]

    def __init__(self, requests_mock: responses.RequestsMock):
        self._valid_auth_codes = []
        self.setup_authorization_endpoint(requests_mock)
        self.setup_token_endpoint(requests_mock)

    def setup_authorization_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            request_url = furl(request.url)
            auth_request = AuthenticationRequest.parse_x_www_form_urlencoded(
                request_url.query.encode()
            )

            code = rand_str()
            self._valid_auth_codes.append(code)
            response = AuthenticationSuccessResponse(
                code=code, state=auth_request.state
            )

            return 301, {"Location": response.encode_url(auth_request.redirect_uri)}, ""

        requests_mock.add_callback(
            method=responses.GET,
            url=self.endpoints["authorization"],
            callback=callback,
        )

    def setup_token_endpoint(self, requests_mock: responses.RequestsMock):
        def callback(request: PreparedRequest) -> Tuple[int, Dict[str, str], str]:
            request_msg = TokenRequest.parse_x_www_form_urlencoded(request.body)
            if request_msg.code == self.cheat_code:
                return (
                    200,
                    {},
                    TokenSuccessResponse(
                        access_token=rand_str(),
                        token_type="Bearer",
                        id_token=rand_str(),
                    ).json(),
                )
            elif request_msg.code in self._valid_auth_codes:
                self._valid_auth_codes.remove(request_msg.code)
                return (
                    200,
                    {},
                    TokenSuccessResponse(
                        access_token=rand_str(),
                        token_type="Bearer",
                        id_token=rand_str(),
                    ).json(),
                )
            else:
                return (
                    400,
                    {},
                    TokenErrorResponse(
                        error=TokenErrorResponse.ErrorType.invalid_grant,
                        error_description="token is invalid",
                    ).json(),
                )

        requests_mock.add_callback(
            method=responses.POST,
            url=self.endpoints["token"],
            callback=callback,
            content_type="application/json",
        )


@pytest.fixture
def dummy_openid_provider(
    mocked_responses: responses.RequestsMock,
) -> DummyOpenidProvider:
    yield DummyOpenidProvider(mocked_responses)


class DummyUserAgent:
    def naviagte_to(self, url: str) -> requests.Response:
        """
        Mimic navigating to the given URL
        """
        return requests.get(url, allow_redirects=True)


@pytest.fixture
def dummy_ua() -> DummyUserAgent:
    yield DummyUserAgent()


class DummyAppServer:
    callback_url = "https://app.example.com/auth/callback"

    def __init__(self, requests_mock: responses.RequestsMock):
        requests_mock.get(
            url=self.callback_url,
            status=200,
            body="dumm app server callback response",
        )


@pytest.fixture
def dummy_app_server(
    mocked_responses: responses.RequestsMock, dummy_openid_provider: DummyOpenidProvider
) -> DummyAppServer:
    return DummyAppServer(mocked_responses)


@pytest.fixture
def mocked_responses():
    """
    A context manager that mocks and de-mocks http request responses
    """
    with responses.RequestsMock(assert_all_requests_are_fired=False) as mock:
        yield mock


@pytest.fixture
def mock_known_provider_configs(mocked_responses: responses.RequestsMock):
    """
    Mock requests to known OpenID provider config URLs to return static content
    """
    # invalid example.com
    mocked_responses.get(
        url="https://example.com/not-found/.well-known/openid-configuration",
        content_type="text/plain",
        body="Issuer config not found",
        status=404,
    )
    mocked_responses.get(
        url="https://example.com/invalid-json-document/.well-known/openid-configuration",
        content_type="application/json",
        body="{ 'hello': 'world' }",
    )
    mocked_responses.get(
        url="https://example.com/invalid-json-syntax/.well-known/openid-configuration",
        content_type="application/json",
        body="{ 'hello }",
    )

    # valid example.com
    mocked_responses.get(
        url="https://provider.example.com/openid-connect/.well-known/openid-configuration",
        json=ProviderMetadata(
            issuer="https://provider.example.com/openid-connect",
            authorization_endpoint="https://provider.example.com/openid-connect/client_auth",
            token_endpoint="https://provider.example.com/openid-connect/token",
            jwks_uri="https://provider.example.com/openid-connect/jwks",
            subject_types_supported=["public"],
            id_token_signing_alg_values_supported=["RS256"],
        ).dict(exclude_defaults=True),
    )
    mocked_responses.get(
        url="https://provider.example.com/openid-connect/jwks",
        json={"keys": []},
    )

    # mafiasi identity
    mocked_responses.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://identity.mafiasi.de/auth/realms/simple_openid_test","authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth","token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","end_session_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/logout","frontchannel_logout_session_supported":true,"frontchannel_logout_supported":true,"jwks_uri":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs","check_session_iframe":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/login-status-iframe.html","grant_types_supported":["authorization_code","implicit","refresh_token","password","client_credentials","urn:ietf:params:oauth:grant-type:device_code","urn:openid:params:grant-type:ciba"],"response_types_supported":["code","none","id_token","token","id_token token","code id_token","code token","code id_token token"],"subject_types_supported":["public","pairwise"],"id_token_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"id_token_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"id_token_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"userinfo_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"request_object_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"response_modes_supported":["query","fragment","form_post","query.jwt","fragment.jwt","form_post.jwt","jwt"],"registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","token_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"token_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"introspection_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"introspection_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"authorization_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"claims_supported":["aud","sub","iss","auth_time","name","given_name","family_name","preferred_username","email","acr"],"claim_types_supported":["normal"],"claims_parameter_supported":true,"scopes_supported":["openid","address","web-origins","offline_access","profile","email","roles","phone","microprofile-jwt"],"request_parameter_supported":true,"request_uri_parameter_supported":true,"require_request_uri_registration":true,"code_challenge_methods_supported":["plain","S256"],"tls_client_certificate_bound_access_tokens":true,"revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","revocation_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"revocation_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"backchannel_logout_supported":true,"backchannel_logout_session_supported":true,"device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","backchannel_token_delivery_modes_supported":["poll","ping"],"backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth","backchannel_authentication_request_signing_alg_values_supported":["PS384","ES384","RS384","ES256","RS256","ES512","PS256","PS512","RS512"],"require_pushed_authorization_requests":false,"pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","mtls_endpoint_aliases":{"token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth"}}',
    )

    # google
    mocked_responses.get(
        "https://accounts.google.com/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://accounts.google.com","authorization_endpoint":"https://accounts.google.com/o/oauth2/v2/auth","device_authorization_endpoint":"https://oauth2.googleapis.com/device/code","token_endpoint":"https://oauth2.googleapis.com/token","userinfo_endpoint":"https://openidconnect.googleapis.com/v1/userinfo","revocation_endpoint":"https://oauth2.googleapis.com/revoke","jwks_uri":"https://www.googleapis.com/oauth2/v3/certs","response_types_supported":["code","token","id_token","code token","code id_token","token id_token","code token id_token","none"],"subject_types_supported":["public"],"id_token_signing_alg_values_supported":["RS256"],"scopes_supported":["openid","email","profile"],"token_endpoint_auth_methods_supported":["client_secret_post","client_secret_basic"],"claims_supported":["aud","email","email_verified","exp","family_name","given_name","iat","iss","locale","name","picture","sub"],"code_challenge_methods_supported":["plain","S256"],"grant_types_supported":["authorization_code","refresh_token","urn:ietf:params:oauth:grant-type:device_code","urn:ietf:params:oauth:grant-type:jwt-bearer"]}',
    )


@pytest.fixture
def mock_known_provider_jwks(mocked_responses: responses.RequestsMock):
    mocked_responses.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs",
        content_type="application/json",
        body='{"keys":[{"kid":"JAehgk0O0uzTEht7KGCPVB_urwfsBGe22phHVDZezeo","kty":"RSA","alg":"RS256","use":"sig","n":"n1t4R6mVlBZggtmEM5HohZYg5QjLtjlRVyAMMM93H4WRPpP2Xqj3RrGwP-LZr6lQhk6w8nvPNGnFujWtXb_dBrhZDBh2z0mCNvfjJMp5Ki5sLT9nCbuZ4NkIhfx7qTzQ--GomQ6uKOSuoE12A_r9UopwuCf_1zrx0MIqN_kMeYU2px5yJW-5qbwPL9hcWcfY_Pdz5SFcQ3kDi6GlXA0PN74Ivhs9baVIv60cwB0TytQXGn__GYFZ-K5qAjIymyIy8cPRy9N27Gs29cWF-C56n1pAxwkcPY1SAQgcrspIhntQbZBeGboUx2pnQ_v8LEqVt9ahh52eAbV36YoQqu1DbQ","e":"AQAB","x5c":["MIICszCCAZsCBgGEZpqrijANBgkqhkiG9w0BAQsFADAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwHhcNMjIxMTExMTIxMDIzWhcNMzIxMTExMTIxMjAzWjAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCfW3hHqZWUFmCC2YQzkeiFliDlCMu2OVFXIAwwz3cfhZE+k/ZeqPdGsbA/4tmvqVCGTrDye880acW6Na1dv90GuFkMGHbPSYI29+MkynkqLmwtP2cJu5ng2QiF/HupPND74aiZDq4o5K6gTXYD+v1SinC4J//XOvHQwio3+Qx5hTanHnIlb7mpvA8v2FxZx9j893PlIVxDeQOLoaVcDQ83vgi+Gz1tpUi/rRzAHRPK1Bcaf/8ZgVn4rmoCMjKbIjLxw9HL03bsazb1xYX4LnqfWkDHCRw9jVIBCByuykiGe1BtkF4ZuhTHamdD+/wsSpW31qGHnZ4BtXfpihCq7UNtAgMBAAEwDQYJKoZIhvcNAQELBQADggEBAD3aKq2xiLNPlUcmxwtR1FihkEM+NM7VV+m+74Lf9m/lhNzuM00JZAW67gIz2kO5K4XaU84PCRzt3sELma/xVoBs3iW8AhtldTax17g0NZzpMEExzYi0vyihq8xihbx+8XrVwxQ3JBkb/y3b+otDX3xRxBKSCDuxuigL1bX7UzgJ3oe4wD6m8yN7nrMdqSMbspd8lkceKNPziclXnbKPIUwrnxdWMmB41zYcT3y+zHxqk6EDkf3NU+C3gCoRCQoBKcK4X2dwZdUbv0ginU9kROjNXc2+gP53CYDT4fqOEX/jBuljbYb9ziaQ9/5kJZ9SsazgWNjpf4VvhAsH64i7+dE="],"x5t":"PxLqu4M4uNWaUhbfkvI0tGrkhtQ","x5t#S256":"OWVSYSSNFaLyfxV0Lg6XyVDnB2t6sHwtNbzVB2fNCVc"}]}',
    )



"""
Refresh token exchange implementation.
"""

import logging
from typing import Union

import requests

from simple_openid_connect.client_authentication import ClientAuthenticationMethod
from simple_openid_connect.data import (
    TokenErrorResponse,
    TokenRequest,
    TokenSuccessResponse,
)

logger = logging.getLogger(__name__)


def exchange_refresh_token(
    token_endpoint: str,
    refresh_token: str,
    client_authentication: ClientAuthenticationMethod,
) -> Union[TokenSuccessResponse, TokenErrorResponse]:
    """
    Exchange a refresh token for new tokens

    :param token_endpoint: The endpoint of the OP at which tokens can be exchanged.
        Corresponds to :data:`ProviderMetadata.token_endpoint <simple_openid_connect.data.ProviderMetadata.token_endpoint>`
    :param refresh_token: The refresh token to use
    :param client_authentication: A way for the client to authenticate itself
    """
    logger.debug("exchanging refresh token for new tokens")
    request_msg = TokenRequest(
        grant_type="refresh_token",
        refresh_token=refresh_token,
        client_id=client_authentication.client_id,
    )
    response = requests.post(
        token_endpoint,
        data=request_msg.encode_x_www_form_urlencoded(),
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        auth=client_authentication,
    )

    if response.status_code == 200:
        response_msg = TokenSuccessResponse.parse_raw(response.content)
        return response_msg
    else:
        response_msg = TokenErrorResponse.parse_raw(response.content)
        return response_msg



"""
`OAuth 2.0 Token Introspection <https://www.rfc-editor.org/rfc/rfc7662>`_ implementation.
"""

from typing import Union

import requests

from simple_openid_connect.client_authentication import ClientAuthenticationMethod
from simple_openid_connect.data import (
    TokenIntrospectionErrorResponse,
    TokenIntrospectionRequest,
    TokenIntrospectionSuccessResponse,
)


def introspect_token(
    introspection_endpoint: str,
    token: str,
    auth: ClientAuthenticationMethod,
    token_type_hint: str = None,
) -> Union[TokenIntrospectionSuccessResponse, TokenIntrospectionErrorResponse]:
    """
    Introspect the given token at the OP.

    :param introspection_endpoint: The token introspection endpoint of the OP.
    :param token: The token to introspect.
    :param auth: Method with which this request is authenticated to the OP.
    :param token_type_hint: Which type of token this is e.g. `refresh_token` or `access_token`.
    :return: The OPs response
    """
    request = TokenIntrospectionRequest(token=token, token_type_hint=token_type_hint)
    response = requests.post(
        introspection_endpoint,
        request.encode_x_www_form_urlencoded(),
        auth=auth,
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
    )

    if response.status_code == 200:
        return TokenIntrospectionSuccessResponse.parse_raw(response.content)
    else:
        return TokenIntrospectionErrorResponse.parse_raw(response.content)



"""
A more contiguous client implementation of the Openid-Connect protocol that offers simpler APIs at the cost of losing some flexibility.
"""

from typing import List, Literal, Optional, Type, TypeVar, Union

from cryptojwt import JWK

from simple_openid_connect import (
    jwk,
    rp_initiated_logout,
    token_introspection,
    token_refresh,
    userinfo,
)
from simple_openid_connect.client_authentication import (
    ClientAuthenticationMethod,
    ClientSecretBasicAuth,
    NoneAuth,
)
from simple_openid_connect.data import (
    IdToken,
    ProviderMetadata,
    RpInitiatedLogoutRequest,
    TokenErrorResponse,
    TokenIntrospectionErrorResponse,
    TokenIntrospectionSuccessResponse,
    TokenSuccessResponse,
    UserinfoErrorResponse,
    UserinfoSuccessResponse,
)
from simple_openid_connect.discovery import discover_configuration_from_issuer
from simple_openid_connect.flows.authorization_code_flow.client import (
    AuthorizationCodeFlowClient,
)

Self = TypeVar("Self", bound="OpenidClient")


class OpenidClient:
    """
    A more contiguous client implementation of the Openid-Connect protocol that offers simpler APIs at the cost of losing some flexibility.
    """

    provider_config: ProviderMetadata
    provider_keys: List[JWK]
    client_auth: ClientAuthenticationMethod
    scope: str

    authorization_code_flow: AuthorizationCodeFlowClient
    "*authorization code flow* related functionality"

    def __init__(
        self,
        provider_config: ProviderMetadata,
        provider_keys: List[JWK],
        authentication_redirect_uri: str,
        client_id: str,
        client_secret: Optional[str] = None,
        scope: str = "openid",
    ):
        self.provider_config = provider_config
        self.provider_keys = provider_keys
        self.authorization_code_flow = AuthorizationCodeFlowClient(self)
        self.scope = scope
        self.authentication_redirect_uri = authentication_redirect_uri

        if client_secret is None:
            self.client_auth = NoneAuth(client_id)
        else:
            if (
                ClientSecretBasicAuth.NAME
                in provider_config.token_endpoint_auth_methods_supported
            ):
                self.client_auth = ClientSecretBasicAuth(client_id, client_secret)
            else:
                raise NotImplementedError(
                    f"a client secret was given but the issuer does not support client_secret_basic authentication which is the only supported method"
                )

    @classmethod
    def from_issuer_url(
        cls: Type[Self],
        url: str,
        authentication_redirect_uri: str,
        client_id: str,
        client_secret: str = None,
        scope: str = "openid",
    ) -> Self:
        """
        Create a new client instance with an issuer url as base, automatically discovering information about the issuer in the process.

        :param url: The url to an Openid issuer
        :param authentication_redirect_uri: URI that is used during the authentication flow to redirect back to this application.
        :param client_id: The already known client id of your application.
            It must be already registered with the issuer.
        :param client_secret: Optionally a client secret which has been assigned to your client from the issuer.
            If not supplied, this client is assumed to be *public* which means it has not client secret because it cannot be kept safe (e.g. a web-app).
        :param scope: Which scopes to request from the OP
        """

        config = discover_configuration_from_issuer(url)
        return cls.from_issuer_config(
            config, authentication_redirect_uri, client_id, client_secret, scope
        )

    @classmethod
    def from_issuer_config(
        cls: Type[Self],
        config: ProviderMetadata,
        authentication_redirect_uri: str,
        client_id: str,
        client_secret: str = None,
        scope: str = "openid",
    ) -> Self:
        """
        Create a new client instance with a resolved issuer configuration as base.

        If you don't have a configuration, use :func:`from_issuer_url` to automatically retrieve it.

        :param config: The configuration of the used issuer.
        :param authentication_redirect_uri: URI that is used during the authentication flow to redirect back to this application.
        :param client_id: The already known client id of your application.
            It must be already registered with the issuer.
        :param client_secret: Optionally a client secret which has been assigned to your client from the issuer.
            If not supplied, this client is assumed to be *public* which means it has not client secret because it cannot be kept safe (e.g. a web-app).
        :param scope: Which scopes to request from the OP
        """
        keys = jwk.fetch_jwks(config.jwks_uri)
        return cls(
            config, keys, authentication_redirect_uri, client_id, client_secret, scope
        )

    @property
    def client_type(self) -> Literal["public", "confidential"]:
        """
        Which type of client behavior is used.

        This is based on whether a client secret has been passed during client construction
        """
        if isinstance(self.client_auth, NoneAuth):
            return "public"
        else:
            return "confidential"

    def fetch_userinfo(
        self, access_token: str
    ) -> Union[UserinfoSuccessResponse, UserinfoErrorResponse]:
        """
        Fetch user information from the OP by doing a userinfo request.

        Which users information is fetched is determined by the OP directly from the used access token.

        :param access_token: An access token which grants access to user information.
        :return: The OPs response
        """
        return userinfo.fetch_userinfo(
            self.provider_config.userinfo_endpoint, access_token
        )

    def decode_id_token(self, raw_token: str) -> IdToken:
        """
        Decode and verify an encoded and signed id token

        :param raw_token: The encoded and signed id token.
            This could e.g. be retrieved as part of the authentication process and returned by the OP in :data:`TokenSuccessResponse.id_token <simple_openid_connect.flows.authorization_code_flow.data.TokenSuccessResponse.id_token>`.
        """
        return IdToken.parse_jwt(raw_token, self.provider_keys)

    def exchange_refresh_token(
        self, refresh_token: str
    ) -> Union[TokenSuccessResponse, TokenErrorResponse]:
        """
        Exchange a refresh token for new tokens

        :param refresh_token: The refresh token to use
        """
        return token_refresh.exchange_refresh_token(
            token_endpoint=self.provider_config.token_endpoint,
            refresh_token=refresh_token,
            client_authentication=self.client_auth,
        )

    def initiate_logout(self, request: RpInitiatedLogoutRequest = None) -> str:
        """
        Initiate user logout as a Relying-Party

        :param request: Additional data pertaining to the logout

        :returns: A url to which the user should be redirected
        """
        return rp_initiated_logout.initiate_logout(
            self.provider_config.end_session_endpoint, request
        )

    def introspect_token(
        self, token: str, token_type_hint: str = None
    ) -> Union[TokenIntrospectionSuccessResponse, TokenIntrospectionErrorResponse]:
        """
        Introspect the given token at the OP.

        :param token: The token to introspect.
        :param token_type_hint: Which type of token this is e.g. `refresh_token` or `access_token`.
        :return: The OPs response
        """
        return token_introspection.introspect_token(
            introspection_endpoint=self.provider_config.introspection_endpoint,
            token=token,
            auth=self.client_auth,
            token_type_hint=token_type_hint,
        )



"""
Internal utilities
"""
import cgi


def is_application_json(content_type: str) -> bool:
    """
    Whether the given content type is `application/json`.
    This is needed because mime types can contain additional options which are ignored here.
    """
    main_type, _options = cgi.parse_header(content_type)
    return main_type == "application/json"



"""
Authentication can follow one of three paths: the Authorization Code Flow, the Implicit Flow (not currently supported), or the Hybrid Flow (not currently supported).
The flows determine how the ID Token and Access Token are returned to the Relying Party (client).

For details see `Section 3 of OpenID Connect Core 1.0 <https://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
"""



import sys
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Callable, Mapping, Tuple

import pytest
import yaml
from furl import furl


@pytest.fixture
def secrets() -> Mapping[str, str]:
    path = Path(__file__).parent / "secrets.yml"
    with open(path, "r", encoding="UTF-8") as f:
        return yaml.safe_load(f)


class RealAppServer(HTTPServer):
    login_url = "http://127.0.0.1:8000/login"
    callback_url = "http://127.0.0.1:8000/callback"

    _on_login = None
    _on_callback = None

    _is_done = False

    def __init__(self):
        super().__init__(
            ("127.0.0.1", 8000), self.RequestHandler, bind_and_activate=True
        )

    def serve_until_done(
        self,
        on_login: Callable[[furl], Tuple[int, Mapping[str, str], str]],
        on_callback: Callable[[furl], Tuple[int, Mapping[str, str], str]],
    ):
        self._on_login = on_login
        self._on_callback = on_callback
        while not self._is_done:
            self.handle_request()

    def done(self):
        self._is_done = True

    def handle_error(self, request, client_address) -> None:
        # re-raise exceptions so that they can fail the test
        raise sys.exc_info()[1]

    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            url = furl(self.path)
            if url.path == "/login":
                response = self.server._on_login(url)
            elif url.path == "/callback":
                response = self.server._on_callback(url)
            else:
                response = 404, {}, "Not found"

            self.send_response(response[0])
            for k, v in response[1].items():
                self.send_header(k, v)
            self.end_headers()
            self.wfile.write(response[2].encode("UTF-8"))


@pytest.fixture
def real_app_server() -> RealAppServer:
    server = RealAppServer()
    yield server
    server.server_close()



"""
Simple and opinionated OpenID-Connect relying party (client) implementation


Protocol Overview
-----------------

The OpenID Connect protocol, in abstract, follows the following steps.

1. The Relying Party (RP / Client) sends a request to the OpenID Provider (OP).
2. The OP authenticates the End-User and obtains authorization.
3. The OP responds with an ID Token and usually an Access Token.
4. The RP can send a request with the Access Token to the UserInfo Endpoint.
5. The UserInfo Endpoint returns Claims about the End-User.

These steps are illustrated in the following diagram::

    +--------+                                   +--------+
    |        |                                   |        |
    |        |---------(1) AuthN Request-------->|        |
    |        |                                   |        |
    |        |  +--------+                       |        |
    |        |  |        |                       |        |
    |        |  |  End-  |<--(2) AuthN & AuthZ-->|        |
    |        |  |  User  |                       |        |
    |   RP   |  |        |                       |   OP   |
    |        |  +--------+                       |        |
    |        |                                   |        |
    |        |<--------(3) AuthN Response--------|        |
    |        |                                   |        |
    |        |---------(4) UserInfo Request----->|        |
    |        |                                   |        |
    |        |<--------(5) UserInfo Response-----|        |
    |        |                                   |        |
    +--------+                                   +--------+
"""
__version__ = "0.0.1"



from simple_openid_connect.client import OpenidClient


def test_full_authorization_code_flow(
    mock_known_provider_configs, dummy_openid_provider, dummy_ua, dummy_app_server
):
    # act
    client = OpenidClient.from_issuer_url(
        url="https://provider.example.com/openid-connect",
        authentication_redirect_uri=dummy_app_server.callback_url,
        client_id=dummy_openid_provider.test_client_id,
        client_secret=dummy_openid_provider.test_client_secret,
    )
    response = dummy_ua.naviagte_to(
        client.authorization_code_flow.start_authentication()
    )
    result = client.authorization_code_flow.handle_authentication_result(response.url)

    # assert
    assert result.access_token
    assert result.id_token



"""
Utilities and data types used when authenticating the relying party (client).

This should not be confused with authenticating a user as the user authentication at the OP is of no concern to us.
Instead, these authenticate the relying party when e.g. exchanging tokens or requesting user information.
It is a necessary process because while access tokens grant a relying party access to certain user information, the
relying party must still identify itself to the OP when accessing that information.

For more information visit `Section 9 of OpenID Connect Core 1.0 <https://openid.net/specs/openid-connect-core-1_0.html#ClientAuthentication>`_ and `Section 2.3 of OAuth 2.0 [RFC6749] <https://www.rfc-editor.org/rfc/rfc6749#section-2.3>`_.
"""

import abc

from requests import models
from requests.auth import AuthBase, HTTPBasicAuth

# TODO Implement more client authentication methods


class ClientAuthenticationMethod(AuthBase, metaclass=abc.ABCMeta):
    """
    A base class for client authentication methods which describes the interface that each method implements to authenticate requests.

    This class also extends requests :class:`AuthBase` so that all derived implementations can directly be used with requests to authenticate something.
    """

    NAME: str
    "How this authentication method is called in the Openid spec"

    @property
    @abc.abstractmethod
    def client_id(self) -> str:
        """
        The client id which is assigned to this app
        """
        raise NotImplementedError()


class NoneAuth(ClientAuthenticationMethod):
    """
    The Client does not authenticate itself at the Token Endpoint, either because it uses only the Implicit Flow (and so does not use the Token Endpoint) or because it is a Public Client with no Client Secret or other authentication mechanism.
    """

    NAME = "none"

    def __init__(self, client_id: str):
        self._client_id = client_id

    def __call__(self, r: models.PreparedRequest) -> models.PreparedRequest:
        return r

    @property
    def client_id(self) -> str:
        return self._client_id


class ClientSecretBasicAuth(ClientAuthenticationMethod, HTTPBasicAuth):
    """
    Clients that have received a `client_secret` value from the Authorization Server authenticate with the Authorization Server using the HTTP Basic authentication scheme.
    """

    NAME = "client_secret_basic"

    @property
    def client_id(self) -> str:
        return self.username

    def __init__(self, client_id: str, client_secret: str):
        """
        :param client_id: The client id which was set or issued during client registration
        :param client_secret: The client secret which was issued during client registration
        """
        super().__init__(username=client_id, password=client_secret)


class AccessTokenBearerAuth(AuthBase):
    """
    Authenticate requests using a given bearer token
    """

    access_token: str

    def __init__(self, access_token: str):
        super().__init__()
        self.access_token = access_token

    def __eq__(self, other):
        return self.access_token == getattr(other, "access_token", None)

    def __ne__(self, other):
        return not self == other

    def __call__(self, r: models.PreparedRequest) -> models.PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self.access_token}"
        return r



"""
The Authorization Code Flow returns an Authorization Code to the Client, which can then exchange it for an ID Token and an Access Token.
This provides the benefit of not exposing any tokens to the User Agent and possibly other malicious applications with access to the User Agent.
The Authorization Server can also authenticate the Client before exchanging the Authorization Code for an Access Token.

**The Authorization Code flow is suitable for Clients that can securely maintain a Client Secret between themselves and the Authorization Server.**
"""
import copy
import logging
from typing import Literal, Union

import requests
from furl import furl

from simple_openid_connect.client_authentication import ClientAuthenticationMethod
from simple_openid_connect.data import (
    AuthenticationErrorResponse,
    AuthenticationRequest,
    AuthenticationSuccessResponse,
    TokenErrorResponse,
    TokenRequest,
    TokenSuccessResponse,
)
from simple_openid_connect.exceptions import AuthenticationFailedError

logger = logging.getLogger(__name__)


def start_authentication(
    authorization_endpoint: str, scope: str, client_id: str, redirect_uri: str
) -> str:
    """
    Start the authentication process by constructing an appropriate :class:`AuthenticationRequest`, serializing it and
    returning a which the end user now needs to visit.

    :returns: A URL to which the user agent should be redirected
    """
    request = AuthenticationRequest(
        scope=scope,
        client_id=client_id,
        redirect_uri=redirect_uri,
        response_type="code",
    )
    return request.encode_url(authorization_endpoint)


def handle_authentication_result(
    current_url: str,
    token_endpoint: str,
    client_authentication: ClientAuthenticationMethod,
    redirect_uri: Union[Literal["auto"], str] = "auto",
) -> Union[TokenSuccessResponse, TokenErrorResponse]:
    """
    Handle an authentication result that is communicated to the RP in form of the user agents current url after having started an authentication process via :func:`start_authentication`.

    :param current_url: The current URL which the user is visiting.
        The authentication result should be encoded into this url by the authorization server.
    :param token_endpoint: The endpoint of the OP at which tokens can be exchanged.
        Corresponds to :data:`ProviderMetadata.token_endpoint <simple_openid_connect.data.ProviderMetadata.token_endpoint>`
    :param client_authentication: A way for the client to authenticate itself
    :param redirect_uri: The `redirect_uri` that was specified during the authentication initiation.
        If the special value `auto` is used, it is assumed that `current_url` is the that callback and it is stripped of query parameters and fragments to reproduce the originally supplied one.

    :raises AuthenticationFailedError: If the current url indicates an authentication failure that prevents an access token from being retrieved.

    :returns: The result of the token exchange
    """
    current_url = furl(current_url)
    if "error" in current_url.query.params.keys():
        auth_response_msg = AuthenticationErrorResponse.parse_url(str(current_url))
        raise AuthenticationFailedError(auth_response_msg)

    if redirect_uri == "auto":
        redirect_uri = str(copy.deepcopy(current_url).remove(fragment=True, query=True))
        logger.debug(
            f"a redirect_uri value of 'auto' was specified. Reproducing redirect_uri (%s) from current_url (%s)",
            redirect_uri,
            current_url,
        )

    auth_response_msg = AuthenticationSuccessResponse.parse_url(str(current_url))
    return exchange_code_for_tokens(
        token_endpoint=token_endpoint,
        authentication_response=auth_response_msg,
        redirect_uri=redirect_uri,
        client_authentication=client_authentication,
    )


def exchange_code_for_tokens(
    token_endpoint: str,
    authentication_response: AuthenticationSuccessResponse,
    redirect_uri: str,
    client_authentication: ClientAuthenticationMethod,
) -> Union[TokenSuccessResponse, TokenErrorResponse]:
    """
    Exchange a received code for access, refresh and id tokens.

    You might want to use :func:`handle_authentication_result` if you don't want to parse an authentication result from the users current url yourself.

    :param token_endpoint: The endpoint of the OP at which tokens can be exchanged.
        Corresponds to :data:`ProviderMetadata.token_endpoint <simple_openid_connect.data.ProviderMetadata.token_endpoint>`
    :param authentication_response: The (successful) response which this app received after the user has come back from
        the OP.
    :param redirect_uri: The callback URI that was specified during the authentication initiation.
    :param client_authentication: A way for the client to authenticate itself

    :returns: The result of the token exchange
    """
    logger.debug("exchanging authentication code for tokens")
    request_msg = TokenRequest(
        code=authentication_response.code,
        redirect_uri=redirect_uri,
        client_id=client_authentication.client_id,
        grant_type="authorization_code",
    )
    response = requests.post(
        token_endpoint,
        data=request_msg.encode_x_www_form_urlencoded(),
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        auth=client_authentication,
    )

    if response.status_code == 200:
        response_msg = TokenSuccessResponse.parse_raw(response.content)
        return response_msg
    else:
        response_msg = TokenErrorResponse.parse_raw(response.content)
        return response_msg



"""
Implementation of Relying-Party initiated logout
"""

from simple_openid_connect.data import RpInitiatedLogoutRequest


def initiate_logout(
    logout_endpoint: str, request: RpInitiatedLogoutRequest = None
) -> str:
    """
    Initiate user logout as a Relying-Party

    :param logout_endpoint: The OPs `end_session_endpoint`
    :param request: Additional data pertaining to the logout

    :returns: A url to which the user should be redirected
    """
    if request is not None:
        return request.encode_url(logout_endpoint)
    else:
        return logout_endpoint
